﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using SalmonViewer;
using System.IO;
using CsGL.OpenGL;
using CsGL.Util;


namespace gl
{
    public partial class EditScene : Form
    {

        public List<Model> models;
        public OGL ogl1;
        
        public EditScene()
        {
            InitializeComponent();
        }

   

        private void bt_add_Click(object sender, EventArgs e)   {
            
            Model model;

            OpenFileDialog od = new OpenFileDialog();
            od.ShowDialog();

            if (od.FileName == "") return;

            if (Path.GetExtension(od.FileName).ToLower() == ".3ds")
            {


                try
                {
                    // Load our 3DS model from the command line argument
                    model = new ThreeDSFile(od.FileName).Model;
                    model.title = Path.GetFileName(od.FileName);
                    models.Add(model);
                    updateListBox();
                    ogl1.PlotGL();
                }
                catch (Exception ex)
                {
                    Console.WriteLine("An Error occured: " + ex.Message);
                }
            }
            else{
                    Console.WriteLine("Not a supported file type.");
                }
 
        }

        private void pxs_Scroll(object sender, EventArgs e)
        {

        }

        private void EditScene_Load(object sender, EventArgs e)
        {
            updateListBox();

        }

        void updateListBox(){
            listBoxModel.Items.Clear();
            
            for(int i = 0; i< models.Count;i++){
                listBoxModel.Items.Add(models[i].title);
            }


            updateSelection();
        }

        void updateSelection() {

            Model m;
            if (listBoxModel.SelectedIndex >= 0)
            {
                m = models[listBoxModel.SelectedIndex];

                pxn.Enabled =
                pyn.Enabled =
                pzn.Enabled =
                rxn.Enabled =
                ryn.Enabled =
                rzn.Enabled =
                sn.Enabled = true;

                pxn.Value = (int)(m.posX * 100);
                pyn.Value = (int)(m.posY * 100);
                pzn.Value = (int)(m.posZ * 100);
                rxn.Value = (int)m.rotX;
                ryn.Value = (int)m.rotY;
                rzn.Value = (int)m.rotZ;
                sn.Value = (int)(m.scale * 100);


            }
            else
                pxn.Enabled =
                pyn.Enabled =
                pzn.Enabled =
                rxn.Enabled =
                ryn.Enabled =
                rzn.Enabled =
                sn.Enabled = false;
        
        }

        private void bt_rem_Click(object sender, EventArgs e)
        {
            if (listBoxModel.SelectedIndex >= 0)
            {
                models.RemoveAt(listBoxModel.SelectedIndex);
                updateListBox();
                ogl1.PlotGL();
            }
        }

        private void listBoxModel_SelectedIndexChanged(object sender, EventArgs e)
        {
            updateSelection();

        }

        private void pxn_ValueChanged(object sender, EventArgs e)
        {
            int i = listBoxModel.SelectedIndex;
            if (i < 0) return;
            models[i].posX = (double)pxn.Value / 100; ;

            ogl1.PlotGL();
        }

        private void pyn_ValueChanged(object sender, EventArgs e)
        {
            int i = listBoxModel.SelectedIndex;
            if (i < 0) return;
            models[i].posY = (double)pyn.Value / 100; ;

            ogl1.PlotGL();
        }

        private void pzn_ValueChanged(object sender, EventArgs e)
        {
        int i = listBoxModel.SelectedIndex;
            if (i < 0) return;
            models[i].posZ = (double)pzn.Value / 100; ;

            ogl1.PlotGL();
        }

        private void sn_ValueChanged(object sender, EventArgs e)
        {
            int i = listBoxModel.SelectedIndex;
            if (i < 0) return;
            models[i].scale = (double)sn.Value / 100; ;

            ogl1.PlotGL();
        }

        private void rxn_ValueChanged(object sender, EventArgs e)
        {
            int i = listBoxModel.SelectedIndex;
            if (i < 0) return;
            models[i].rotX  = (double)rxn.Value  ;

            ogl1.PlotGL();
        }

        private void ryn_ValueChanged(object sender, EventArgs e)
        {
            int i = listBoxModel.SelectedIndex;
            if (i < 0) return;
            models[i].rotY = (double)ryn.Value ;

            ogl1.PlotGL();
        }

        private void rzn_ValueChanged(object sender, EventArgs e)
        {
            int i = listBoxModel.SelectedIndex;
            if (i < 0) return;
            models[i].rotZ = (double)rzn.Value ;

            ogl1.PlotGL();
        }

       

        private void lz_ValueChanged(object sender, EventArgs e)
        {
            ogl1.luz0posicao[2] = (float)lz.Value / 100;
            ogl1.PlotGL();
        }

        private void ly_ValueChanged(object sender, EventArgs e)
        {
            ogl1.luz0posicao[1] = (float)ly.Value / 100;
            ogl1.PlotGL();

        }

        private void lx_ValueChanged(object sender, EventArgs e)
        {
            ogl1.luz0posicao[0] = (float)lx.Value / 100;
            ogl1.PlotGL();

        }

        private void bt_ambient_Click(object sender, EventArgs e)
        {
            colorDialog1.ShowDialog();

            ogl1.luz0ambiente2[0] = (float)colorDialog1.Color.R / 256;
            ogl1.luz0ambiente2[1] = (float)colorDialog1.Color.G / 256;
            ogl1.luz0ambiente2[2] = (float)colorDialog1.Color.B / 256;

            bt_ambient.BackColor = colorDialog1.Color;
            ogl1.PlotGL();
        }

        private void bt_difuse_Click(object sender, EventArgs e)
        {
            colorDialog1.ShowDialog();

            ogl1.luz0difusa[0] = (float)colorDialog1.Color.R / 256;
            ogl1.luz0difusa[1] = (float)colorDialog1.Color.G / 256;
            ogl1.luz0difusa[2] = (float)colorDialog1.Color.B / 256;
            bt_difuse.BackColor = colorDialog1.Color;
            ogl1.PlotGL();
        }

        private void label11_Click(object sender, EventArgs e)
        {

        }

        private void bt_getdif_Click(object sender, EventArgs e)
        {

          
            ogl1.luz0difusa[0] = ogl1.average[0] * 2;
            ogl1.luz0difusa[1] = ogl1.average[1] * 2;
            ogl1.luz0difusa[2] = ogl1.average[2] * 2;

            if (ogl1.luz0difusa[0] > 1) ogl1.luz0difusa[0] = 1;
            if (ogl1.luz0difusa[1] > 1) ogl1.luz0difusa[1] = 1;
            if (ogl1.luz0difusa[2] > 1) ogl1.luz0difusa[2] = 1;  

            bt_difuse.BackColor = Color.FromArgb(
                (int)(ogl1.average[0]  * 256),
                (int)(ogl1.average[1]  * 256),
                (int)(ogl1.average[2]  * 256));

         
            ogl1.PlotGL();
        }

        private void bt_getamb_Click(object sender, EventArgs e)
        {

            ogl1.luz0ambiente2[0] = ogl1.average[0];
            ogl1.luz0ambiente2[1] = ogl1.average[1];
            ogl1.luz0ambiente2[2] = ogl1.average[2];


            bt_ambient.BackColor = Color.FromArgb(
                (int)(ogl1.average[0] * 256),
                (int)(ogl1.average[1] * 256),
                (int)(ogl1.average[2] * 256));

            ogl1.PlotGL();
        }

        private void button1_Click(object sender, EventArgs e)
        {
            Hide();
        }
     
    }
}
