﻿using CsGL.OpenGL;
using System;
using System.Drawing;
using System.Windows.Forms;
using CsGL.Util;
using System.Collections.Generic;
using System.Collections;

using SalmonViewer;


//using Triangulator.Geometry;
namespace gl
{
    #region OpenGL Control

    public class OGL : OpenGLControl
    {
        public class Vector3f
        {
            public float x, y, z;

            public static void cross(Vector3f Result, Vector3f v1, Vector3f v2)
            {
                Result.x = (v1.y * v2.z) - (v1.z * v2.y);
                Result.y = (v1.z * v2.x) - (v1.x * v2.z);
                Result.z = (v1.x * v2.y) - (v1.y * v2.x);
            }

            public static float dot(Vector3f v1, Vector3f v2)
            {
                return (v1.x * v2.x) + (v1.y * v2.y) + (v1.z + v2.z);
            }

            public virtual float length()
            {
                return (float)System.Math.Sqrt(x * x + y * y + z * z);
            }
        }

        #region variable definitions

        private static int height = 0;
        private static int width = 0;

        private System.Object matrixLock = new System.Object();
        private float[] matrix = new float[16];

        private static bool isLeftDrag = false;
        private static bool isRightDrag = false;

        public uint[] texture = new uint[10];

        public int distance = 5;

        // Camer Properties
        float zoom = 1;//aproximação do objeto em relação a camera

        private float anglex_i, angley_i; // usados para calcular a angulação
        public float anglex = 0; //abgulação no eixo x
        public float angley = 0; //angulação no eixo y

        public float panx = 0;
        public float pany = 0;

        private float panx_i = 0;
        private float pany_i = 0;

        public bool wire = false;
        public bool shadows = true;

        public double fov = Math.PI /6;
        double alpha;


        //Object Models
        public List<Model> models;

        //mesh scenario
        OpenGLTexture2D Texture;

        public List<Triangulator.Geometry.Triangle> tris = new List<Triangulator.Geometry.Triangle>();
        public List<Triangulator.Geometry.Point> points = new List<Triangulator.Geometry.Point>();
        public List<Triangulator.Geometry.Point> txpoints = new List<Triangulator.Geometry.Point>();

        //luzes
        private static float[] Minv = new float[16];
        public float[] luz0difusa = { 0.8f, 0.8f, 0.8f, 1.0f };
        public float[] luz0ambiente2 = { 0.23f, 0.3f, 0.223f, 1.0f };
        public float[] luz0espec = { 0.0f, 0.0f, 0.0f, 0.0f };
        public float[] luz0posicao = { 0f, 0f, .0f, 1.0f };

        public float[] average = new float[3];
        #endregion

        #region hide
        public OGL()
            : base()
        {
            this.KeyDown += new KeyEventHandler(OurView_OnKeyDown);
            GC.Collect();
        }

        protected override OpenGLContext CreateContext()
        {
            // Setup Our PixelFormat And Context
            ControlGLContext context = new ControlGLContext(this);
            DisplayType displayType = new DisplayType(32, 16, 2, 0);
            context.Create(displayType, null);
           
            return context;
        }

        protected override void InitGLContext()
        {


            #region glconfig

           // GL.glClearStencil(0);

            GL.glShadeModel(GL.GL_SMOOTH);								// enable smooth shading
            GL.glClearColor(0.0f, 0.0f, 0.0f, 0.5f);					// black background
            GL.glClearDepth(1.0f);										// depth buffer setup
            GL.glEnable(GL.GL_DEPTH_TEST);								// enables depth testing
            GL.glDepthFunc(GL.GL_LEQUAL);								// type of depth testing
            GL.glHint(GL.GL_PERSPECTIVE_CORRECTION_HINT, GL.GL_NICEST);	// nice perspective calculations
            GL.glPointSize(4.0f);

            GL.gluPerspective(25.0, (double)width / (double)height, 1.0, 15.0);
            GL.glTranslatef(0.0f, 0.0f, -4.0f);


            GL.glCullFace(GL.GL_BACK);														// Set Culling Face To Back Face
            GL.glEnable(GL.GL_CULL_FACE);

   
            #endregion


            #region mouse handles
            MouseControl mouseControl = new MouseControl(this);
            mouseControl.AddControl(this);
            mouseControl.LeftMouseDown += new MouseEventHandler(glOnLeftMouseDown);
            mouseControl.LeftMouseUp += new MouseEventHandler(glOnLeftMouseUp);

            mouseControl.RightMouseDown += new MouseEventHandler(glOnRightMouseDown);
            mouseControl.RightMouseUp += new MouseEventHandler(glOnRightMouseUp);
            #endregion

            //camera
            alpha = Math.Tan(fov / 2);

        }

        void setCamera()
        {

            GL.glMatrixMode(GL.GL_PROJECTION);
            GL.glLoadIdentity();

            GL.glFrustum(-alpha / 2, alpha / 2, -alpha / 2, alpha / 2, .5, 1000.0); // substituir por d


            GL.glTranslated(panx / 1000, -pany / 1000, 0);
            GL.glRotated(180, 0, 1, 0);

            //roda a canera no eixo x e y	


            GL.glTranslated(0, 0, 1);

            GL.glRotated(anglex, 0, 1, 0);
            GL.glRotated(angley, Math.Cos(anglex * Math.PI / 180), 0, Math.Sin(anglex * Math.PI / 180));

            GL.glTranslated(0, 0, -1);



            GL.glMatrixMode(GL.GL_MODELVIEW);
            GL.glLoadIdentity();
        }

        protected override void OnSizeChanged(EventArgs e)
        {
            base.OnSizeChanged(e);
            Size s = Size;
            height = s.Height;
            width = s.Width;

            GL.glViewport(0, 0, width, height);

            setCamera();

            //arcBall.setBounds((float)width, (float)height); //*NEW* Update mouse bounds for arcball
        }

        //textura
        public void updateTexture(string filename)
        {

            GL.glEnable(GL.GL_TEXTURE_2D);
            Bitmap image = new Bitmap(filename);


            //image.RotateFlip(RotateFlipType.RotateNoneFlipY);
            System.Drawing.Imaging.BitmapData bitmapdata;
            Rectangle rect = new Rectangle(0, 0, image.Width, image.Height);

            bitmapdata = image.LockBits(rect, System.Drawing.Imaging.ImageLockMode.ReadOnly,
                 image.PixelFormat);
            //GL

            GL.glGenTextures(4, texture);
            GL.glBindTexture(GL.GL_TEXTURE_2D, texture[0]);
            GL.glTexImage2D(GL.GL_TEXTURE_2D, 0, (int)GL.GL_RGB8, image.Width, image.Height,
                0, GL.GL_BGR_EXT, GL.GL_UNSIGNED_BYTE, bitmapdata.Scan0);
            GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_LINEAR);		// Linear Filtering
            GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_LINEAR);		// Linear Filtering

            image.UnlockBits(bitmapdata);
            image.Dispose();

        }
        #endregion

        public void PlotGL()
        {
            try
            {
                GL.glColor3f(0, 1, 1);
                GL.glEnable(GL.GL_CULL_FACE );
                GL.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT | GL.GL_STENCIL_BUFFER_BIT); // Clear screen and DepthBuffer
                GL.glLoadIdentity();

                //  GL.glClearStencil(0); //clear the stencil buffer
                // GL.glClearDepth(1.0f);

                plotScene();

                plotModel();
                               
                GL.glFlush();     // Flush the GL Rendering Pipeline

                this.Invalidate();

            }
            catch
            {
                return;
            }

        }

        //draw shadows on scene
        void plotShadow(Model m)
        {


            float[] lp = localLight(m, luz0posicao);

            setVisibility(m, lp);

            GL.glColorMask((byte)GL.GL_FALSE, (byte)GL.GL_FALSE, (byte)GL.GL_FALSE, (byte)GL.GL_FALSE);
            GL.glDepthMask((byte)GL.GL_FALSE);

            GL.glEnable(GL.GL_STENCIL_TEST);

            GL.glCullFace(GL.GL_FRONT);
            GL.glStencilFunc(GL.GL_ALWAYS, 0x0, 0xff);
            GL.glStencilOp(GL.GL_KEEP, GL.GL_INCR, GL.GL_KEEP);

            plotModelShadowVolume(m,lp);
            
            GL.glCullFace(GL.GL_BACK);
            GL.glStencilFunc(GL.GL_ALWAYS, 0x0, 0xff);
            GL.glStencilOp(GL.GL_KEEP, GL.GL_DECR, GL.GL_KEEP);

            plotModelShadowVolume(m,lp);

            //GL.glStencilFunc(GL.GL_GREATER , 0x0, 0xff); 
    
            m.Render();

           
        }

        void setVisibility(Model m, float[] lp)
        {
            float side;

            if (!m.plane)
                CalcPlane(m);

            foreach (Entity el in m.Entities)
            {
                // Set Visual Parameter
                for (int i = 0; i < el.indices.Length; i++)
                {
                    // Check To See If Light Is In Front Or Behind The Plane (Face Plane)
                    side =
                        el.indices[i].a * lp[0] +
                        el.indices[i].b * lp[1] +
                        el.indices[i].c * lp[2] +
                        el.indices[i].d * 1;//lp[3] ;

                    if (side > 0)
                        el.indices[i].visible = true;
                    else
                        el.indices[i].visible = false;
                    
                }
            }
        }
        //darken the shadows

        void darken()
        {
            GL.glDisable(GL.GL_LIGHTING);
            GL.glMatrixMode(GL.GL_MODELVIEW );
            GL.glPushMatrix();
            GL.glLoadIdentity();
            GL.glMatrixMode(GL.GL_PROJECTION);
            GL.glPushMatrix();
            GL.glLoadIdentity();
            GL.glOrtho(0, 1, 1, 0, 0, 1);
            GL.glDisable(GL.GL_DEPTH_TEST);

            GL.glEnable(GL.GL_BLEND);
            GL.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);

            GL.glColor4f(0.0f, 0.0f, 0.0f, 0.2f);
            GL.glBegin(GL.GL_QUADS);
            GL.glVertex2i(0, 0);
            GL.glVertex2i(0, 1);
            GL.glVertex2i(1, 1);
            GL.glVertex2i(1, 0);
            GL.glEnd();

            GL.glDisable(GL.GL_BLEND);
            GL.glEnable(GL.GL_DEPTH_TEST);
            GL.glPopMatrix();
            GL.glMatrixMode(GL.GL_MODELVIEW);
            GL.glPopMatrix();
        }

        //draw shadow vomumes of model
        void plotModelShadowVolume(Model m,float[] light){



            float infty =10;
            
            int k = 0;
            int a = 0;
            foreach (Entity el in m.Entities)
            {
                foreach (Triangle t in el.indices)
                {
                   a++;
                   if (t.visible){
                        #region prism
                        GL.glBegin(GL.GL_QUADS);

                        //1st face
                        GL.glVertex3f(
                            (float)el.vertices[t.Vertex1].X,
                            (float)el.vertices[t.Vertex1].Y,
                            (float)el.vertices[t.Vertex1].Z);

                        GL.glVertex3f(
                            (float)(el.vertices[t.Vertex1].X - light[0]) * infty + light[0],
                            (float)(el.vertices[t.Vertex1].Y - light[1]) * infty + light[1],
                            (float)(el.vertices[t.Vertex1].Z - light[2]) * infty + light[2]);

                        GL.glVertex3f(
                            (float)(el.vertices[t.Vertex2].X - light[0]) * infty + light[0],
                            (float)(el.vertices[t.Vertex2].Y - light[1]) * infty + light[1],
                            (float)(el.vertices[t.Vertex2].Z - light[2]) * infty + light[2]);

                        GL.glVertex3f(
                            (float)el.vertices[t.Vertex2].X,
                            (float)el.vertices[t.Vertex2].Y,
                            (float)el.vertices[t.Vertex2].Z);
                        //2nd face
                        GL.glVertex3f(
                            (float)el.vertices[t.Vertex2].X,
                            (float)el.vertices[t.Vertex2].Y,
                            (float)el.vertices[t.Vertex2].Z);

                        GL.glVertex3f(
                            (float)(el.vertices[t.Vertex2].X - light[0]) * infty + light[0],
                            (float)(el.vertices[t.Vertex2].Y - light[1]) * infty + light[1],
                            (float)(el.vertices[t.Vertex2].Z - light[2]) * infty + light[2]);

                        GL.glVertex3f(
                            (float)(el.vertices[t.Vertex3].X - light[0]) * infty + light[0],
                            (float)(el.vertices[t.Vertex3].Y - light[1]) * infty + light[1],
                            (float)(el.vertices[t.Vertex3].Z - light[2]) * infty + light[2]);

                        GL.glVertex3f(
                            (float)el.vertices[t.Vertex3].X,
                            (float)el.vertices[t.Vertex3].Y,
                            (float)el.vertices[t.Vertex3].Z);

                        //3rd face
                        GL.glVertex3f(
                            (float)el.vertices[t.Vertex3].X,
                            (float)el.vertices[t.Vertex3].Y,
                            (float)el.vertices[t.Vertex3].Z);

                        GL.glVertex3f(
                            (float)(el.vertices[t.Vertex3].X - light[0]) * infty + light[0],
                            (float)(el.vertices[t.Vertex3].Y - light[1]) * infty + light[1],
                            (float)(el.vertices[t.Vertex3].Z - light[2]) * infty + light[2]);

                        GL.glVertex3f(
                            (float)(el.vertices[t.Vertex1].X - light[0]) * infty + light[0],
                            (float)(el.vertices[t.Vertex1].Y - light[1]) * infty + light[1],
                            (float)(el.vertices[t.Vertex1].Z - light[2]) * infty + light[2]);
                        GL.glVertex3f(
                            (float)el.vertices[t.Vertex1].X,
                            (float)el.vertices[t.Vertex1].Y,
                            (float)el.vertices[t.Vertex1].Z);


                        GL.glEnd();
                        #endregion
                        k++;
                        //if (k == 50)
                        //    break;
                    }
                   
                   
                }

            }
        }

        //altera as coordenadas da luz para local em 1 objeto
        float[] localLight(Model m, float[] light0) {
            float[] light = { light0[0], light0[1], light0[2], 1 };
            GL.glPushMatrix();
            GL.glLoadIdentity();

            //translada o objeto

            //scala o objeto
            GL.glScaled(1/m.scale, 1/(m.scale * 1.3), 1/m.scale);

            //roda o objeto
            GL.glRotated(-m.rotZ, 0, 0, 1);
            GL.glRotated(-m.rotY, 0, 1, 0);
            GL.glRotated(-m.rotX, 1, 0, 0);

            GL.glTranslated(-m.posX, -m.posY, -(m.posZ + 1));


            GL.glGetFloatv(GL.GL_MODELVIEW_MATRIX, Minv);

            VMatMult(Minv, ref light);

            GL.glPopMatrix();

            return light;
        }

        //draw scenario
        private void plotScene()
        {
            #region solid
            //float[] luz0ambiente = { 1.0f, 1.0f, 1.0f, 1.0f };

            GL.glDisable(GL.GL_LIGHTING);

            GL.glEnable(GL.GL_TEXTURE_2D);
            GL.glBindTexture(GL.GL_TEXTURE_2D, texture[0]);
            //GL.glBindTexture(GL.GL_TEXTURE_2D, texture[0]);

            GL.glColor3f(1,1,1);

            GL.glBegin(GL.GL_TRIANGLES);
            foreach (Triangulator.Geometry.Triangle t in tris)
            {
                GL.glTexCoord2d(txpoints[t.p1].X, txpoints[t.p1].Y);
                GL.glVertex3d(
                    points[t.p1].X,
                    points[t.p1].Y,
                    points[t.p1].Z);

                GL.glTexCoord2d(txpoints[t.p2].X, txpoints[t.p2].Y);
                GL.glVertex3d(
                    points[t.p2].X,
                    points[t.p2].Y,
                    points[t.p2].Z);

                GL.glTexCoord2d(txpoints[t.p3].X, txpoints[t.p3].Y);
                GL.glVertex3d(
                    points[t.p3].X,
                    points[t.p3].Y,
                    points[t.p3].Z);
            }

            GL.glEnd();



            GL.glDisable(GL.GL_TEXTURE_2D);
            #endregion

            #region wireframe

            if (wire)
            {

                GL.glColor3f(1,1,1);
                GL.glBegin(GL.GL_LINES);                                    // Start A Triangle Strip

                foreach (Triangulator.Geometry.Triangle t in tris)
                {

                    GL.glVertex3d(
                        points[t.p1].X,
                        points[t.p1].Y,
                        points[t.p1].Z);

                    GL.glVertex3d(
                        points[t.p2].X,
                        points[t.p2].Y,
                        points[t.p2].Z);

                    GL.glVertex3d(
                        points[t.p2].X,
                        points[t.p2].Y,
                        points[t.p2].Z);

                    GL.glVertex3d(
                        points[t.p3].X,
                        points[t.p3].Y,
                        points[t.p3].Z);

                    GL.glVertex3d(
                        points[t.p3].X,
                        points[t.p3].Y,
                        points[t.p3].Z);

                    GL.glVertex3d(
                        points[t.p1].X,
                        points[t.p1].Y,
                        points[t.p1].Z);

                }

                GL.glEnd();
            }
            #endregion
        }

        //draw all models
        void plotModel(){


            if (shadows)
            {
                GL.glClear(GL.GL_STENCIL_BUFFER_BIT);
                foreach (Model m in models)
                {


                    GL.glPushMatrix();
                    //translada o objeto
                    GL.glTranslated(m.posX, m.posY, m.posZ + 1);

                    //scala o objeto
                    GL.glScaled(m.scale, m.scale * 1.3, m.scale);

                    //roda o objeto
                    GL.glRotated(m.rotX, 1, 0, 0);
                    GL.glRotated(m.rotY, 0, 1, 0);
                    GL.glRotated(m.rotZ, 0, 0, 1);

                    //desenha as sombras do objeto

                    plotShadow(m);

                    GL.glPopMatrix();
                }

                GL.glDisable(GL.GL_POLYGON_OFFSET_FILL);
                GL.glDisable(GL.GL_CULL_FACE);
                GL.glColorMask((byte)GL.GL_TRUE, (byte)GL.GL_TRUE, (byte)GL.GL_TRUE, (byte)GL.GL_TRUE);
                GL.glDepthMask((byte)GL.GL_TRUE);

                GL.glStencilFunc(GL.GL_NOTEQUAL, 0x0, 0xff);
                GL.glStencilOp(GL.GL_REPLACE, GL.GL_REPLACE, GL.GL_REPLACE);
                darken();

                GL.glDisable(GL.GL_STENCIL_TEST);
            }


            GL.glEnable(GL.GL_LIGHTING);
            foreach (Model m in models)
            {
                //limpa o stencil utilizado para fazer sombras
                GL.glPushMatrix();
                //translada o objeto
                GL.glTranslated(m.posX, m.posY, m.posZ + 1);

                //scala o objeto
                GL.glScaled(m.scale, m.scale * 1.3, m.scale);

                //roda o objeto
                GL.glRotated(m.rotX, 1, 0, 0);
                GL.glRotated(m.rotY, 0, 1, 0);
                GL.glRotated(m.rotZ, 0, 0, 1);

                //configura a luz
                setLight();

                //desenha o objeto
                m.Render();


                GL.glPopMatrix();
               


            }
            // Enable Depth Mask

        }

        //set light
        void setLight() {

            // lights
            GL.glMatrixMode(GL.GL_MODELVIEW);
            GL.glPushMatrix();
            GL.glLoadIdentity();

            GL.glEnable(GL.GL_LIGHTING);
            GL.glEnable(GL.GL_LIGHT0);

            GL.glLightfv(GL.GL_LIGHT0, GL.GL_POSITION, luz0posicao);
            GL.glLightfv(GL.GL_LIGHT0, GL.GL_AMBIENT, luz0ambiente2);
            GL.glLightfv(GL.GL_LIGHT0, GL.GL_DIFFUSE, luz0difusa);
            GL.glLightfv(GL.GL_LIGHT0, GL.GL_SPECULAR, luz0espec);
            GL.glPopMatrix();
        }
        #region CalcPlane
        /// <summary>
        ///  Computes a plane equation given 3 points.
        /// </summary>
        /// <param name="o">The GlObject.</param>
        /// <param name="plane">The plane to store the calculations into.</param>
        private void CalcPlane(Model m)
        {

            Vector3f [] v = new Vector3f [4];
            v[1] = new Vector3f();
            v[2] = new Vector3f();
            v[3] = new Vector3f();
            foreach (Entity el in m.Entities)
            {

                for (int i = 0; i < el.indices.Length; i++)
                {
                    Triangle plane = el.indices[i];

                    v[1].x = (float)el.vertices[el.indices[i].Vertex1].X;
                    v[1].y = (float)el.vertices[el.indices[i].Vertex1].Y;
                    v[1].z = (float)el.vertices[el.indices[i].Vertex1].Z;

                    v[2].x = (float)el.vertices[el.indices[i].Vertex2].X;
                    v[2].y = (float)el.vertices[el.indices[i].Vertex2].Y;
                    v[2].z = (float)el.vertices[el.indices[i].Vertex2].Z;

                    v[3].x = (float)el.vertices[el.indices[i].Vertex3].X;
                    v[3].y = (float)el.vertices[el.indices[i].Vertex3].Y;
                    v[3].z = (float)el.vertices[el.indices[i].Vertex3].Z;


                    el.indices[i].a = v[1].y * (v[2].z - v[3].z) + v[2].y * (v[3].z - v[1].z) + v[3].y * (v[1].z - v[2].z);
                    el.indices[i].b = v[1].z * (v[2].x - v[3].x) + v[2].z * (v[3].x - v[1].x) + v[3].z * (v[1].x - v[2].x);
                    el.indices[i].c = v[1].x * (v[2].y - v[3].y) + v[2].x * (v[3].y - v[1].y) + v[3].x * (v[1].y - v[2].y);
                    el.indices[i].d = -(v[1].x * (v[2].y * v[3].z - v[3].y * v[2].z) + v[2].x * (v[3].y * v[1].z - v[1].y * v[3].z) + v[3].x * (v[1].y * v[2].z - v[2].y * v[1].z));
                }
            }
            m.plane = true;
        }
        #endregion

        private void CastShadow(Model m, float[] lp)
        {

            uint i, j, jj;
            int p1, p2;
            Vector3f v1 = new Vector3f();
            Vector3f v2 = new Vector3f();

            //float side;
            Entity el = m.Entities[0];

            GL.glDisable(GL.GL_LIGHTING);
            GL.glDepthMask((byte)GL.GL_FALSE);
            GL.glDepthFunc(GL.GL_LEQUAL);

            GL.glEnable(GL.GL_STENCIL_TEST);
            GL.glColorMask(0, 0, 0, 0);
            GL.glStencilFunc(GL.GL_ALWAYS, 1, 0xffffffff);

            // First Pass, Stencil Operation Decreases Stencil Value
            GL.glFrontFace(GL.GL_CCW);
            GL.glStencilOp(GL.GL_KEEP, GL.GL_KEEP, GL.GL_INCR);

            for (i = 0; i < el.indices.Length; i++)
            {
                if (true)//el.indices[i].visible)
                {
                    for (j = 0; j < 3; j++)
                    {
                        //k = el.indices[i].neigh[j];
                        //if ((k <= 0) || (!el.indices[k - 1].visible))                       {
                        // Here We Have An Edge, We Must Draw A Polygon
                        int[] vertices = { el.indices[i].Vertex1, el.indices[i].Vertex2, el.indices[i].Vertex3 };

                        p1 = vertices[j];
                        jj = (j + 1) % 3;
                        p2 = vertices[jj];

                        // Calculate The Length Of The Vector
                        v1.x = (float)(el.vertices[p1].X - lp[0]) * 10;
                        v1.y = (float)(el.vertices[p1].Y - lp[1]) * 10;
                        v1.z = (float)(el.vertices[p1].Z - lp[2]) * 10;

                        v2.x = (float)(el.vertices[p2].X - lp[0]) * 10;
                        v2.y = (float)(el.vertices[p2].Y - lp[1]) * 10;
                        v2.z = (float)(el.vertices[p2].Z - lp[2]) * 10;

                        // Draw The Polygon

                        GL.glBegin(GL.GL_TRIANGLE_STRIP);
                        GL.glVertex3f((float)el.vertices[p1].X, (float)el.vertices[p1].Y, (float)el.vertices[p1].Z);
                        GL.glVertex3f(
                            (float)(el.vertices[p1].X + v1.x),
                            (float)(el.vertices[p1].Y + v1.y),
                            (float)(el.vertices[p1].Z + v1.z));
                        GL.glVertex3f(
                            (float)el.vertices[p2].X,
                            (float)el.vertices[p2].Y,
                            (float)el.vertices[p2].Z);
                        GL.glVertex3f(
                            (float)(el.vertices[p2].X + v2.x),
                            (float)(el.vertices[p2].Y + v2.y),
                            (float)(el.vertices[p2].Z + v2.z));
                        GL.glEnd();
                        // }
                    }
                }
            }

            // Second Pass, Stencil Operation Increases Stencil Value
            GL.glFrontFace(GL.GL_CW);
            GL.glStencilOp(GL.GL_KEEP, GL.GL_KEEP, GL.GL_DECR);

            for (i = 0; i < 0; i++)//el.indices.Length; i++)
            {
                if (el.indices[i].visible)
                {
                    
                    for (j = 0; j < 3; j++)
                    {
                        // k = el.indices[i].neigh[j];
                        // if ((k <= 0) || (!el.indices[k - 1].visible))    {
                        // Here We Have An Edge, We Must Draw A Polygon

                        int[] vertices = { el.indices[i].Vertex1, el.indices[i].Vertex2, el.indices[i].Vertex3 };

                        p1 = vertices[j];
                        jj = (j + 1) % 3;
                        p2 = vertices[jj];

                        // Calculate The Length Of The Vector
                        v1.x = (float)(el.vertices[p1].X - lp[0]) * 10;
                        v1.y = (float)(el.vertices[p1].Y - lp[1]) * 10;
                        v1.z = (float)(el.vertices[p1].Z - lp[2]) * 10;

                        v2.x = (float)(el.vertices[p2].X - lp[0]) * 10;
                        v2.y = (float)(el.vertices[p2].Y - lp[1]) * 10;
                        v2.z = (float)(el.vertices[p2].Z - lp[2]) * 10;

                        // Draw The Polygon
                        GL.glBegin(GL.GL_TRIANGLE_STRIP);
                        GL.glVertex3f(
                            (float)(el.vertices[p1].X),
                            (float)(el.vertices[p1].Y),
                            (float)(el.vertices[p1].Z));
                        GL.glVertex3f(
                            (float)(el.vertices[p1].X + v1.x),
                            (float)(el.vertices[p1].Y + v1.y),
                            (float)(el.vertices[p1].Z + v1.z));
                        GL.glVertex3f(
                            (float)(el.vertices[p2].X),
                            (float)(el.vertices[p2].Y),
                            (float)(el.vertices[p2].Z));
                        GL.glVertex3f(
                            (float)(el.vertices[p2].X + v2.x),
                            (float)(el.vertices[p2].Y + v2.y),
                            (float)(el.vertices[p2].Z + v2.z));
                        GL.glEnd();
                    }
                    //}
                }
            }

            GL.glFrontFace(GL.GL_CCW);
            GL.glColorMask(1, 1, 1, 1);

            // Draw A Shadowing Rectangle Covering The Entire Screen
            GL.glColor4f(0.0f, 0.0f, 0.0f, 0.4f);
            GL.glEnable(GL.GL_BLEND);
            GL.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
            GL.glStencilFunc(GL.GL_NOTEQUAL, 0, 0xffffffff);
            GL.glStencilOp(GL.GL_KEEP, GL.GL_KEEP, GL.GL_KEEP);
            GL.glPushMatrix();
            GL.glLoadIdentity();
            GL.glBegin(GL.GL_TRIANGLE_STRIP);
            GL.glVertex3f(-0.1f, 0.1f, -0.10f);
            GL.glVertex3f(-0.1f, -0.1f, -0.10f);
            GL.glVertex3f(0.1f, 0.1f, -0.10f);
            GL.glVertex3f(0.1f, -0.1f, -0.10f);
            GL.glEnd();
            GL.glPopMatrix();
            GL.glDisable(GL.GL_BLEND);

            GL.glDepthFunc(GL.GL_LEQUAL);
            GL.glDepthMask((byte)GL.GL_TRUE);
            GL.glEnable(GL.GL_LIGHTING);
            GL.glDisable(GL.GL_STENCIL_TEST);
            GL.glShadeModel(GL.GL_SMOOTH);
        }

        #region mesh control

        public void updateGeometry(List<Triangulator.Geometry.Point> pts,
                            List<Triangulator.Geometry.Triangle> tri)
        {
            points.Clear();
            txpoints.Clear();
            tris.Clear();
            foreach (Triangulator.Geometry.Point p in pts)
            {
                points.Add(new Triangulator.Geometry.Point(p.X, p.Y, p.Z));
                txpoints.Add(new Triangulator.Geometry.Point(p.X, p.Y, p.Z));
            }

            tris = new List<Triangulator.Geometry.Triangle>(tri);
            normalize(points);
            //normalizeUV(txpoints);
            PlotGL();
        }

        private void normalize(List<Triangulator.Geometry.Point> pts)
        {
            double maxz = 0;
            double minz = 0;

            foreach (Triangulator.Geometry.Point p in pts)
            {
                if (p.Z > maxz) maxz = p.Z;
                if (p.Z < minz) minz = p.Z;
            }

            foreach (Triangulator.Geometry.Point p in pts)
            {
                p.Z = ((p.Z - minz) * distance / 10) + 1;
                p.Z = p.Z * p.Z;

                p.X = -(p.X - .5) * 2;
                p.Y = -(p.Y - .5) * 2;

            }

            foreach (Triangulator.Geometry.Point p in pts)
            {
                p.X = p.X * alpha * p.Z;
                p.Y = p.Y * alpha * p.Z;
            }

        }

        #endregion

        #region Keyboard Control
        protected void OurView_OnKeyDown(object Sender, KeyEventArgs kea)
        {
            //if escape was pressed, exit the application
            if (kea.KeyCode == Keys.D)
            {
                panx--;
            }

            //if R was pressed, reset
            if (kea.KeyCode == Keys.A)
            {
                panx++;
            }
            setCamera();
        }
        #endregion Keyboard Control

        #region Mouse Control


        private void drag(Point MousePt)
        {

        }

        public void glOnMouseMove(object sender, MouseEventArgs e)
        {
            if (isLeftDrag)
            {
                panx = e.X + panx_i;
                pany = e.Y + pany_i;
                setCamera();
                PlotGL();
            }
            if (isRightDrag)
            {
                anglex = e.X + anglex_i;
                angley = e.Y + angley_i;
                setCamera();
                PlotGL();
            }



        }

        public void glOnLeftMouseDown(object sender, MouseEventArgs e)
        {
            isLeftDrag = true;
            panx_i = panx - e.X;
            pany_i = pany - e.Y;

        }

        public void glOnLeftMouseUp(object sender, MouseEventArgs e)
        {
            isLeftDrag = false;
        }

        private void glOnRightMouseDown(object sender, MouseEventArgs e)
        {
            isRightDrag = true;
            anglex_i = anglex - e.X;
            angley_i = angley - e.Y;
        }

        private void glOnRightMouseUp(object sender, MouseEventArgs e)
        {
            isRightDrag = false;
            //this.reset();
        }
        #endregion

        #region VMatMult(float[] M, ref float[] v)
        /// <summary>
        /// Multiplies a vector by a matrix.
        /// </summary>
        /// <param name="M">The matrix.</param>
        /// <param name="v">The vector.</param>
        private void VMatMult(float[] M, ref float[] v)
        {
            float[] res = new float[4];													// Hold Calculated Results

            res[0] = M[0] * v[0] + M[4] * v[1] + M[8] * v[2] + M[12] * v[3];			// Multiply A Vector By A Matrix
            res[1] = M[1] * v[0] + M[5] * v[1] + M[9] * v[2] + M[13] * v[3];
            res[2] = M[2] * v[0] + M[6] * v[1] + M[10] * v[2] + M[14] * v[3];
            res[3] = M[3] * v[0] + M[7] * v[1] + M[11] * v[2] + M[15] * v[3];

            v[0] = res[0];																// Results Are Stored Back In v[]
            v[1] = res[1];
            v[2] = res[2];
            v[3] = res[3];																// Homogenous Coordinate
        }
        #endregion VMatMult(float[] M, ref float[] v)

    }
    #endregion OpenGL Control

    #region MouseControl
    public class MouseControl
    {
        protected Control newCtrl;
        protected MouseButtons FinalClick;

        public event EventHandler LeftClick;
        public event EventHandler RightClick;
        public event MouseEventHandler LeftMouseDown;
        public event MouseEventHandler LeftMouseUp;
        public event MouseEventHandler RightMouseDown;
        public event MouseEventHandler RightMouseUp;

        public Control Control
        {
            get { return newCtrl; }
            set
            {
                newCtrl = value;
                Initialize();
            }
        }

        public MouseControl()
        {
        }

        public MouseControl(Control ctrl)
        {
            Control = ctrl;
        }

        public void AddControl(Control ctrl)
        {
            Control = ctrl;
        }

        protected virtual void Initialize()
        {
            newCtrl.Click += new EventHandler(OnClick);
            newCtrl.MouseDown += new MouseEventHandler(OnMouseDown);
            newCtrl.MouseUp += new MouseEventHandler(OnMouseUp);

        }

        private void OnClick(object sender, EventArgs e)
        {
            switch (FinalClick)
            {
                case MouseButtons.Left:
                    if (LeftClick != null)
                    {
                        LeftClick(sender, e);
                    }
                    break;

                case MouseButtons.Right:
                    if (RightClick != null)
                    {
                        RightClick(sender, e);
                    }
                    break;
            }
        }

        private void OnMouseDown(object sender, MouseEventArgs e)
        {
            FinalClick = e.Button;

            switch (e.Button)
            {
                case MouseButtons.Left:
                    {
                        if (LeftMouseDown != null)
                        {
                            LeftMouseDown(sender, e);
                        }
                        break;
                    }

                case MouseButtons.Right:
                    {
                        if (RightMouseDown != null)
                        {
                            RightMouseDown(sender, e);
                        }
                        break;
                    }
            }
        }

        private void OnMouseUp(object sender, MouseEventArgs e)
        {
            switch (e.Button)
            {
                case MouseButtons.Left:
                    {
                        if (LeftMouseUp != null)
                        {
                            LeftMouseUp(sender, e);
                        }
                        break;
                    }

                case MouseButtons.Right:
                    {
                        if (RightMouseUp != null)
                        {
                            RightMouseUp(sender, e);
                        }
                        break;
                    }
            }
        }
    }
    #endregion MouseControl

}